#cython: boundscheck=False, wraparound=False, initializedcheck=False

import numpy as np
cimport numpy as np
from numpy import ndarray
from numpy cimport ndarray
from numpy.math cimport logaddexp, INFINITY as inf
cdef extern from "math.h" nogil :
    np.float64_t exp(np.float64_t x)

cpdef dict forward(np.ndarray[np.int64_t, ndim=2] lattice, np.ndarray[np.float64_t, ndim=3] x_dot_parameters, long S):
    """ Helper to calculate the forward weights.  """
    cdef dict alpha = {}

    cdef unsigned int r
    cdef unsigned int i0, j0, s0, i1, j1, s1, edge_parameter_index
    cdef unsigned int I, J, s

    cdef unsigned int old_i0, old_j0, old_s0 
    cdef np.float64_t edge_potential

    old_i0, old_j0, old_s0 = -1, -1, -1

    for r in range(lattice.shape[0]):
        i0, j0, s0 = lattice[r, 0], lattice[r, 1], lattice[r, 2] 
        i1, j1, s1 = lattice[r, 3], lattice[r, 4], lattice[r, 5]
        edge_parameter_index = lattice[r, 6]
        
        if i0 != old_i0 or j0 != old_j0 or s0 != old_s0:
            if i0 == 0 and j0 == 0:
                alpha[(i0, j0, s0)] = x_dot_parameters[i0, j0, s0]
            else:
                alpha[(i0, j0, s0)] += x_dot_parameters[i0, j0, s0]

            old_i0, old_j0, old_s0 = i0, j0, s0

        edge_potential = (x_dot_parameters[i1, j1, edge_parameter_index]
                          + <np.float64_t> alpha[(i0, j0, s0)])
        alpha[(i0, j0, s0, i1, j1, s1, edge_parameter_index)] = edge_potential
        alpha[(i1, j1, s1)] = logaddexp(<np.float64_t> alpha.get((i1, j1, s1), -inf),
                                        edge_potential)

    I = x_dot_parameters.shape[0] - 1
    J = x_dot_parameters.shape[1] - 1

    for s in range(S):
        if I == J == 0:
            alpha[(I, J, s)] = x_dot_parameters[I, J, s]
        else:
            alpha[(I, J, s)] = <np.float64_t> alpha.get((I, J, s), -inf) + x_dot_parameters[I, J, s]

    return alpha

cpdef np.float64_t[::1] forward_predict(np.int64_t[:, ::1] lattice,
                                        np.float64_t[:, :, ::1] x_dot_parameters,
                                        long S) :
    """ Helper to calculate the forward weights for prediction.  """

    cdef np.float64_t[:, :, ::1] alpha = x_dot_parameters.copy()
    alpha[:] = -inf

    cdef unsigned int r 
    cdef unsigned int i0, j0, s0, i1, j1, s1, edge_parameter_index

    cdef int old_s0 = -1

    cdef np.float64_t edge_potential, source_node_potential

    for r in range(lattice.shape[0]):
        i0, j0, s0 = lattice[r, 0], lattice[r, 1], lattice[r, 2]

        if s0 != old_s0 :
            if i0 == 0 and j0 == 0:
                source_node_potential = x_dot_parameters[i0, j0, s0]
            else:
                source_node_potential = (alpha[i0,j0,s0]
                                         + x_dot_parameters[i0,j0,s0])
            old_s0 = s0

        i1, j1, s1 = lattice[r, 3], lattice[r, 4], lattice[r, 5]
        edge_parameter_index = lattice[r, 6]

        edge_potential = (x_dot_parameters[i1, j1, edge_parameter_index]
                          + source_node_potential)

        alpha[i1, j1, s1] = logaddexp(alpha[i1, j1, s1], edge_potential)

    cdef int I = alpha.shape[0] - 1
    cdef int J = alpha.shape[1] - 1

    for s in range(S):
        if I == J == 0 :
            alpha[I, J, s] = x_dot_parameters[I, J, s]
        else:
            alpha[I, J, s] += x_dot_parameters[I, J, s]

    cdef np.float64_t[::1] final_alphas = alpha[J - 1, I - 1, :S]
    cdef np.float64_t Z = -inf

    for s in range(S):
        Z = logaddexp(Z, final_alphas[s])

    for s in range(S):
        final_alphas[s] = exp(final_alphas[s] - Z)

    return final_alphas


cpdef np.float64_t[:, :, ::1] forward_max_predict(np.int64_t[:, ::1] lattice,
                                            np.float64_t[:, :, ::1] x_dot_parameters,
                                            long S) :
    """ Helper to calculate the forward max-sum weights for prediction.  """

    cdef np.float64_t[:, :, ::1] alpha = x_dot_parameters.copy()
    alpha[:] = -inf

    cdef unsigned int r
    cdef unsigned int i0, j0, s0, i1, j1, s1, edge_parameter_index

    cdef int old_s0 = -1

    cdef np.float64_t edge_potential, source_node_potential

    for r in range(lattice.shape[0]):
        i0, j0, s0 = lattice[r, 0], lattice[r, 1], lattice[r, 2]

        if s0 != old_s0 :
            if i0 == 0 and j0 == 0:
                source_node_potential = x_dot_parameters[i0, j0, s0]
            else:
                source_node_potential = (alpha[i0,j0,s0]
                                         + x_dot_parameters[i0,j0,s0])
            old_s0 = s0

        i1, j1, s1 = lattice[r, 3], lattice[r, 4], lattice[r, 5]
        edge_parameter_index = lattice[r, 6]

        edge_potential = (x_dot_parameters[i1, j1, edge_parameter_index]
                          + source_node_potential)

        alpha[i1, j1, s1] = max(alpha[i1, j1, s1], edge_potential)

    cdef int I = alpha.shape[0] - 1
    cdef int J = alpha.shape[1] - 1

    for s in range(S):
        if I == J == 0 :
            alpha[I, J, s] = x_dot_parameters[I, J, s]
        else:
            alpha[I, J, s] += x_dot_parameters[I, J, s]

    return alpha


cpdef dict backward(ndarray[np.int64_t, ndim=2] lattice,
                    ndarray[np.float64_t, ndim=3] x_dot_parameters,
                    long I, long J, long S):
    """ Helper to calculate the backward weights.  """
    cdef dict beta = {}

    cdef unsigned int r
    cdef unsigned int s
    cdef unsigned int i0, j0, s0, i1, j1, s1, edge_parameter_index

    cdef np.float64_t edge_potential

    for s in range(S):
        beta[(I-1, J-1, s)] = 0.0

    for r in range((lattice.shape[0] - 1), -1, -1):
        i0, j0, s0 = lattice[r, 0], lattice[r, 1], lattice[r, 2], 
        i1, j1, s1 = lattice[r, 3], lattice[r, 4], lattice[r, 5]
        edge_parameter_index = lattice[r, 6]

        edge_potential = <np.float64_t> beta[(i1, j1, s1)] + x_dot_parameters[i1, j1, s1]
        beta[(i0, j0, s0, i1, j1, s1, edge_parameter_index)] = edge_potential
        beta[(i0, j0, s0)] = logaddexp(<np.float64_t> beta.get((i0, j0, s0), -inf),
                                       (edge_potential 
                                        + x_dot_parameters[i1, 
                                                           j1, 
                                                           edge_parameter_index]))
    return beta


def gradient(dict alpha,
             dict beta,
             ndarray[np.float64_t, ndim=2] parameters,
             ndarray[np.int64_t] states_to_classes,
             ndarray[np.float64_t, ndim=3] x,
             long y,
             long I, long J, long K):
    """ Helper to calculate the marginals and from that the gradient given the forward and backward weights. """
    cdef unsigned int n_classes = max(states_to_classes) + 1
    cdef ndarray[np.float64_t] class_Z = np.zeros((n_classes,))
    cdef np.float64_t Z = -inf
    cdef np.float64_t weight
    cdef unsigned int k

    for state, clas in enumerate(states_to_classes):
        weight = <np.float64_t> alpha[(I - 1, J - 1, state)]
        class_Z[clas] = weight
        Z = logaddexp(Z, weight)

    cdef ndarray[np.float64_t, ndim=2] derivative = np.full_like(parameters, 0.0)
    cdef unsigned int i0, j0, s0, i1, j1, s1, edge_parameter_index
    cdef np.float64_t alphabeta

    for node in alpha.viewkeys() | beta.viewkeys():
        if len(node) == 3:
            i0, j0, s0 = node
            alphabeta = <np.float64_t>alpha[(i0, j0, s0)] + <np.float64_t>beta[(i0, j0, s0)]

            for k in range(K):
                if states_to_classes[s0] == y:
                    derivative[s0, k] += (exp(alphabeta - class_Z[y]) - exp(alphabeta - Z)) * x[i0, j0, k]
                else:
                    derivative[s0, k] -= exp(alphabeta - Z) * x[i0, j0, k]

        else:
            i0, j0, s0, i1, j1, s1, edge_parameter_index = node
            alphabeta = <np.float64_t>alpha[(i0, j0, s0, i1, j1, s1, edge_parameter_index)] \
                        + <np.float64_t>beta[(i0, j0, s0, i1, j1, s1, edge_parameter_index)]

            for k in xrange(K):
                if states_to_classes[s1] == y:
                    derivative[edge_parameter_index, k] += (exp(alphabeta - class_Z[y]) - exp(alphabeta - Z)) * x[i1, j1, k]
                else:
                    derivative[edge_parameter_index, k] -= exp(alphabeta - Z) * x[i1, j1, k]

    return (class_Z[y]) - (Z), derivative


def gradient_sparse(dict alpha,
                    dict beta,
                    ndarray[np.float64_t, ndim=2] parameters,
                    ndarray[np.int64_t] states_to_classes,
                    ndarray[np.int64_t, ndim=3] x_index,
                    ndarray[np.float64_t, ndim=3] x_value,
                    long y,
                    long I, long J, long K):
    """
    Helper to calculate the marginals and from that the gradient given the forward and backward weights, for
    sparse input features.
    """
    cdef unsigned int n_classes = max(states_to_classes) + 1
    cdef ndarray[np.float64_t] class_Z = np.zeros((n_classes,))
    cdef np.float64_t Z = -inf
    cdef np.float64_t weight
    cdef unsigned int C = K
    cdef unsigned int c
    cdef int k

    for state, clas in enumerate(states_to_classes):
        weight = <np.float64_t> alpha[(I - 1, J - 1, state)]
        class_Z[clas] = weight
        Z = logaddexp(Z, weight)

    cdef ndarray[np.float64_t, ndim=2] derivative = np.full_like(parameters, 0.0)
    cdef unsigned int i0, j0, s0, i1, j1, s1, edge_parameter_index
    cdef np.float64_t alphabeta

    for node in alpha.viewkeys() | beta.viewkeys():
        if len(node) == 3:
            i0, j0, s0 = node
            alphabeta = <np.float64_t>alpha[(i0, j0, s0)] + <np.float64_t>beta[(i0, j0, s0)]

            for c in range(C):
                k = x_index[i0, j0, c]
                if k < 0:
                    break
                if states_to_classes[s0] == y:
                    derivative[s0, k] += (exp(alphabeta - class_Z[y]) - exp(alphabeta - Z)) * x_value[i0, j0, c]
                else:
                    derivative[s0, k] -= exp(alphabeta - Z) * x_value[i0, j0, c]

        else:
            i0, j0, s0, i1, j1, s1, edge_parameter_index = node
            alphabeta = <np.float64_t>alpha[(i0, j0, s0, i1, j1, s1, edge_parameter_index)] \
                                + <np.float64_t>beta[(i0, j0, s0, i1, j1, s1, edge_parameter_index)]

            for c in range(C):
                k = x_index[i1, j1, c]
                if k < 0:
                    break
                if states_to_classes[s1] == y:
                    derivative[edge_parameter_index, k] += (exp(alphabeta - class_Z[y]) - exp(alphabeta - Z)) * x_value[i1, j1, c]
                else:
                    derivative[edge_parameter_index, k] -= exp(alphabeta - Z) * x_value[i1, j1, c]

    return (class_Z[y]) - (Z), derivative


def populate_sparse_features(ndarray[np.float64_t, ndim=3] x,
                             ndarray[np.int64_t, ndim=3] index_array,
                             ndarray[np.float64_t, ndim=3] value_array,
                             long I, long J, long K):
    """ Helper to fill in sparse feature arrays. """
    cdef unsigned int i, j, c, k
    for i in range(I):
        for j in range(J):
            c = 0
            for k in range(K):
                if x[i, j, k] != 0:
                    value_array[i, j, c] = x[i, j, k]
                    index_array[i, j, c] = k
                    c += 1

def sparse_multiply(ndarray[np.float64_t, ndim=3] answer,
                    ndarray[np.int64_t, ndim=3] index_array,
                    ndarray[np.float64_t, ndim=3] value_array,
                    ndarray[np.float64_t, ndim=2] dense_array,
                    long I, long J, long K, long C, long S):
    """ Multiply a sparse three dimensional numpy array (using our own scheme) with a two dimensional array. """
    cdef unsigned int i, j, s, c
    cdef int k
    for i in range(I):
        for j in range(J):
            for s in range(S):
                for c in range(C):
                    k = index_array[i, j, c]
                    if k < 0:
                        break
                    answer[i, j, s] += value_array[i, j, c] * dense_array[k, s]



#cython: boundscheck=False, wraparound=False, initializedcheck=False
#cython: infer_types=True

import numpy as np
cimport numpy as np
from numpy import ndarray
from numpy cimport ndarray
from numpy.math cimport INFINITY as inf
from libc.math cimport exp
cdef extern from "log1p.h" nogil:
    np.float64_t log1p(np.float64_t x)

cdef np.float64_t LOG_2 = 0.6931471805599453
cdef np.float64_t LOG_3 = 1.0986122886681098

cpdef dict forward(np.ndarray[np.float64_t, ndim=3] x_dot_parameters, int S):
    """ Helper to calculate the forward weights.  """
    cdef dict alpha = {}

    cdef int I, J
    I, J = x_dot_parameters.shape[0], x_dot_parameters.shape[1]

    # Fill in the edges of the state matrices
    #
    #   0 1 2 3 
    # 0 x x x x
    # 1 x - - -
    # 2 x - - -
    # 3 x - - -
    cdef int matching =  1 * S
    cdef int deletion =  2 * S
    cdef int insertion = 3 * S
    cdef np.float64_t insert, delete, match
    cdef int i, j, s 

    for s in range(S):
        alpha[0, 0, s] = x_dot_parameters[0, 0, s]
        for i in range(1, I):
            insert = (alpha[i - 1, 0, s] +
                      x_dot_parameters[i, 0, insertion + s])
            alpha[i, 0, s] = x_dot_parameters[i, 0, s] + insert

            alpha[i - 1, 0, s, i, 0, s, insertion + s] = insert
        for j in range(1, J):
            delete = (alpha[0, j - 1, s] +
                      x_dot_parameters[0, j, deletion + s])
            alpha[0, j, s] = x_dot_parameters[0, j, s] + delete

            alpha[0, j - 1, s, 0, j, s, deletion + s] = delete
        
        # Now fill in the middle of the matrix    
        for i in range(1, I):
            for j in range(1, J):
                insert = (alpha[i - 1, j, s] +
                          x_dot_parameters[i, j, insertion + s])
                delete = (alpha[i, j - 1, s] +
                          x_dot_parameters[i, j, deletion + s])
                match = (alpha[i - 1, j - 1, s] +
                         x_dot_parameters[i, j, matching + s])
                alpha[i, j, s] = (x_dot_parameters[i, j, s] +
                                  logsumexp(insert, delete, match))

                alpha[i - 1, j, s, i, j, s, insertion + s] = insert
                alpha[i, j - 1, s, i, j, s, deletion + s] = delete
                alpha[i - 1, j - 1, s, i, j, s, matching + s] = match

    return alpha

cpdef dict backward(np.ndarray[np.float64_t, ndim=3] x_dot_parameters, int S):
    """ Helper to calculate the forward weights.  """
    cdef dict beta = {}

    cdef int I, J
    I, J = x_dot_parameters.shape[0], x_dot_parameters.shape[1]

    # Fill in the edges of the state matrices
    #
    #   0 1 2 3 
    # 0 - - - x
    # 1 - - - x
    # 2 - - - x
    # 3 x x x x
    cdef int matching =  1 * S
    cdef int deletion =  2 * S
    cdef int insertion = 3 * S
    cdef np.float64_t insert, delete, match
    cdef int i, j, s
    cdef int last_row = I - 1
    cdef int last_col = J - 1

    for s in range(S):
        beta[last_row, last_col, s] = 0
        for i in range(last_row - 1, -1, -1):
            insert = (beta[i + 1, last_col, s] +
                      x_dot_parameters[i + 1, last_col, s])
            beta[i, last_col, s] = (x_dot_parameters[i + 1, last_col, insertion + s]
                                    + insert)

            beta[i, last_col, s, i + 1, last_col, s, insertion + s] = insert
        for j in range(last_col - 1, -1, -1):
            delete = (beta[last_row, j + 1, s] +
                      x_dot_parameters[last_row, j + 1, s])
            beta[last_row, j, s] = (x_dot_parameters[last_row, j + 1, deletion + s]
                                    + delete)

            beta[last_row, j, s, last_row, j + 1, s, deletion + s] = delete
        
        # Now fill in the middle of the matrix    
        for i in range(last_row - 1, -1, -1):
            for j in range(last_col - 1, -1, -1):
                insert = (beta[i + 1, j, s] +
                          x_dot_parameters[i + 1, j, s])
                delete = (beta[i, j + 1, s] +
                          x_dot_parameters[i, j + 1, s])
                match = (beta[i + 1, j + 1, s] +
                         x_dot_parameters[i + 1, j + 1, s])

                beta[i, j, s, i + 1, j, s, insertion + s] = insert
                beta[i, j, s, i, j + 1, s, deletion + s] = delete
                beta[i, j, s, i + 1, j + 1, s, matching + s] = match

                insert += x_dot_parameters[i + 1, j, insertion + s]
                delete += x_dot_parameters[i, j + 1, deletion + s]
                match += x_dot_parameters[i + 1, j + 1, matching + s]
                
                beta[i, j, s] = logsumexp(insert, delete, match)

    return beta

cpdef np.float64_t[::1] forward_predict(np.float64_t[:, :, :] x_dot_parameters,
                                              int S) :
    cdef np.float64_t[:, :, ::1] alpha = x_dot_parameters.copy()

    cdef int I, J
    I, J = alpha.shape[0], alpha.shape[1]

    # Fill in the edges of the state matrices
    #
    #   0 1 2 3 
    # 0 x x x x
    # 1 x - - -
    # 2 x - - -
    # 3 x - - -
    cdef int matching =  1 * S
    cdef int deletion =  2 * S
    cdef int insertion = 3 * S
    cdef np.float64_t insert, delete, match
    cdef int i, j, s
    
    for s in range(S):
        alpha[0, 0, s] = x_dot_parameters[0, 0, s]
        for i in range(1, I):
            insert = (alpha[i - 1, 0, s] +
                      x_dot_parameters[i, 0, insertion + s])
            alpha[i, 0, s] = x_dot_parameters[i, 0, s] + insert
        for j in range(1, J):
            delete = (alpha[0, j - 1, s] +
                      x_dot_parameters[0, j, deletion + s])
            alpha[0, j, s] = x_dot_parameters[0, j, s] + delete
        
        # Now fill in the middle of the matrix    
        for j in range(1, J):
            for i in range(1, I):
                insert = (alpha[i - 1, j, s] +
                          x_dot_parameters[i, j, insertion + s])
                delete = (alpha[i, j - 1, s] +
                          x_dot_parameters[i, j, deletion + s])
                match = (alpha[i - 1, j - 1, s] +
                         x_dot_parameters[i, j, matching + s])
                alpha[i, j, s] = (x_dot_parameters[i, j, s] +
                                  logsumexp(insert, delete, match))


    cdef np.float64_t[::1] final_alphas = alpha[I - 1, J - 1, :S]

    cdef np.float64_t Z = -inf
    for s in range(S):
        Z = logaddexp(Z, final_alphas[s])

    for s in range(S):
        final_alphas[s] = exp(final_alphas[s] - Z)

    return final_alphas

cdef np.float64_t logaddexp(np.float64_t x, np.float64_t y) nogil:
    cdef np.float64_t tmp
    if x == y :
        return x + LOG_2
    else :
        tmp = x - y
        if tmp > 0 :
            return x + log1p(exp(-tmp))
        elif tmp <= 0 :
            return y + log1p(exp(tmp))
        else :
            return tmp

cdef np.float64_t logsumexp(np.float64_t x, np.float64_t y, np.float64_t z) nogil :
    if x == y == z:
        return x + LOG_3
    elif x > y > z or x > z > y:
        return x + log1p(exp(y - x) + exp(z - x))
    elif x > y == z:
        return x + log1p(exp(y - x) * 2)
    elif x == y > z:
        return x + LOG_2 + log1p(exp(z - x - LOG_2))
    elif x == z > y:
        return x + LOG_2 + log1p(exp(y - x - LOG_2))
    elif y > x > z or y > z > x:
        return y + log1p(exp(x - y) + exp(z - y))
    elif y > x == z:
        return y + log1p(exp(x - y) * 2)
    elif y == z > x:
        return y + LOG_2 + log1p(exp(x - y - LOG_2))
    elif z > x > y or z > y > x:
        return z + log1p(exp(x - z) + exp(y - z))
    elif z > x == y:
        return z + log1p(exp(x - z) * 2)



from setuptools import setup, Extension
import numpy
import numpy.distutils

try:
    from Cython.Build import cythonize
    use_cython = True
except ImportError:
    use_cython = False

if use_cython:
    ext_modules = cythonize([Extension('pyhacrf.algorithms',
                                       ['pyhacrf/algorithms.pyx'],
                                       extra_compile_args = ["-ffast-math", "-O4"],
                                       **numpy.distutils.misc_util.get_info('npymath')),
                             Extension('pyhacrf.adjacent',
                                       ['pyhacrf/adjacent.pyx'],
                                       include_dirs=[numpy.get_include()],
                                       extra_compile_args = ["-ffast-math", "-O4"])])
else:
    ext_modules = [Extension('pyhacrf.algorithms',
                             ['pyhacrf/algorithms.c'],
                             extra_compile_args = ["-ffast-math", "-O4"],
                             **numpy.distutils.misc_util.get_info('npymath')),
                   Extension('pyhacrf.adjacent',
                             ['pyhacrf/adjacent.c'],
                             include_dirs=[numpy.get_include()],
                             extra_compile_args = ["-ffast-math", "-O4"])]


def readme():
    with open('README.rst') as f:
        return f.read()

setup(
    name='pyhacrf-datamade',
    version='0.2.3',
    packages=['pyhacrf'],
    install_requires=["numpy>=1.10.4; python_version<'3.6'",
                      "numpy>=1.12.1; python_version=='3.6'",
                      "numpy>=1.15.0; python_version=='3.7'",
                      'PyLBFGS>=0.1.3'],
    ext_modules=ext_modules,
    url='https://github.com/datamade/pyhacrf',
    author='Dirko Coetsee',
    author_email='dpcoetsee@gmail.com',
    maintainer='Forest Gregg',
    maintainer_email='fgregg@gmail.com',
    description='Hidden alignment conditional random field, a discriminative string edit distance',
    long_description=readme(),
    classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
        ],
    )
